# Copyright (C) 2008 Ansgar Wiechers <ansgar.wiechers@planetcobalt.net>
#
# This module is free software. You may use, redistribute and/or modify it
# under the terms of the GNU General Public License Version 2.0 as found at
# <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>.

package SQLite;
use FilterDB::DB;
our @ISA = qw( DB );

=head1 NAME

  FilterDB::SQLite -- Backscatter filter backend (SQLite)

=head1 SYNOPSIS

  use FilterDB::SQLite;
  use strict;

  my $db = new SQLite(
      'filename' => '/PATH/TO/my.db',
    );  

  if ($db->check('foo')) {
    print "'foo' found\n";
  } else {
    print "'foo' not found\n";
  }

  $db->insert('foo');
  $db->insert('bar');

  if ($db->check('foo')) {
    print "'foo' found\n";
  } else {
    print "'foo' not found\n";
  }

  my %records = $db->list();
  my @keys = keys %records;
  if ($#keys >= 0) {
    print "$records{$key}: $key\n" foreach my $key (@keys);
  }

=head1 REQUIRES

Perl5.8.8, DBD::SQLite

=head1 DESCRIPTION

The SQLite class implements a SQLite backend for the backscatter filter. It
stores MD5 hashes of given values and the timestamps when the values were
entered into the database. It provides methods to check whether the hash for any
given value is present in the database, to add insert new values, to list the
database contents, and to remove outdated values.

Status information will be logged to syslog if the calling application has
opened the syslog via openlog().

=head1 METHODS

=head2 Creation

=over 4

=item new($filename, [$max_age])

Opens a connection to an SQLite database stored in $filename. If the file isn't
present, the connection attempt will fail. The optional argument $max_age
specifies the number of days after which an entry should be removed from the
database (by the cleanup() function). The default is 4 days.

=back

=head2 Access

=over 4

=item check($value)

Checks whether the MD5 hash of the given $value is present in the database.
Returns TRUE if the hash is present, otherwise FALSE.

=item cleanup()

Removes entries older than $max_age days from the database.

=item insert($value)

Inserts the MD5 hash of the given $value and the current timestamp into the
database. If the hash is already present, the existing entry is updated with the
durrent timestamp.

=item list()

Returns a dictionary with all MD5 hash->timestamp pairs stored in the database.

=back

=head1 AUTHOR

Ansgar Wiechers <ansgar.wiechers@planetcobalt.net>

=cut

require 5.8.8;

use strict;
use warnings;
use DBI;
use Date::Manip;
use Digest::MD5 qw( md5_hex );
use Sys::Syslog qw( :standard :macros );

use constant TRUE	=> 1;
use constant FALSE => not TRUE;

my $driver = 'dbi:SQLite';
my $default_max_age = 4; # days

my $check_statement   = 'SELECT hash FROM hashes WHERE hash = md5(?)';
my $insert_statement  = 'INSERT INTO hashes (hash, created) VALUES (md5(?), now())';
my $update_statement  = 'UPDATE hashes SET created = now() WHERE hash = md5(?)';
my $cleanup_statement = 'DELETE FROM hashes WHERE CAST(age(created) AS integer) > CAST(? AS integer)';
my $list_statement    = 'SELECT hash, created FROM hashes';

sub new {
	my ($class, %args) = @_;
	my $self = bless { %args }, $class;

	$self->{max_age} = $default_max_age unless defined $self->{max_age} and $self->{max_age} =~ m/^[0-9]+$/;

	_terminate("Opening database connection failed: database file '$self->{filename}' does not exist\n") unless -e $self->{filename};
	syslog(LOG_DEBUG, "Opening database connection to '$self->{filename}'");

	$self->{db} = DBI->connect("$driver:dbname=$self->{filename};", '', '', {PrintError => TRUE, AutoCommit => FALSE})
		or _terminate('Opening database connection failed: ' . DBI->errstr . "\n");

	# Define the three functions now(), md5() and age() that are to be used in the
	# prepared statements.
	$self->{db}->func('now', 0, sub { return ParseDate('today'); }, 'create_function');
	$self->{db}->func('md5', 1, sub { return md5_hex($_[0]); }, 'create_function');
	$self->{db}->func('age', 1, sub { return Delta_Format(DateCalc($_[0], 'today'), 0, "%dh"); },'create_function');

	$self->{check_query} = $self->{db}->prepare($check_statement)
		or _terminate("Preparing SELECT statement failed: " . DBI->errstr . "\n");
	$self->{insert_query} = $self->{db}->prepare($insert_statement)
		or _terminate("Preparing INSERT INTO statement failed: " . DBI->errstr . "\n");
	$self->{update_query} = $self->{db}->prepare($update_statement)
		or _terminate("Preparing UPDATE statement failed: " . DBI->errstr . "\n");
	$self->{cleanup_query} = $self->{db}->prepare($cleanup_statement)
		or _terminate("Preparing DELETE statement failed: " . DBI->errstr . "\n");
	$self->{list_query} = $self->{db}->prepare($list_statement)
		or _terminate("Preparing SELECT * statement failed: " . DBI->errstr . "\n");

	return $self;
}

sub DESTROY {
	my ($self) = @_;
	syslog(LOG_DEBUG, "Closing database connection");
	if ($self->{db}) {
		undef $self->{check_query};
		undef $self->{insert_query};
		undef $self->{update_query};
		undef $self->{cleanup_query};
		undef $self->{list_query};
		$self->{db}->rollback;
		$self->{db}->disconnect;
	}
}

sub check {
	my ($self, $value) = @_;
	my $hash;

	_terminate('check(): Missing argument $value') unless defined $value;

	syslog(LOG_DEBUG, "Checking for presence of '$value'");

	$self->{check_query}->execute($value)
		or _terminate("SELECT failed: " . DBI->errstr . "\n");
	$self->{check_query}->bind_columns(\$hash)
		or _terminate("Binding result column failed: " . DBI->errstr . "\n");
	$self->{check_query}->fetchrow_arrayref();
	$self->{check_query}->finish;
	$self->{db}->rollback;

	return ((defined $hash) ? TRUE : FALSE);
}

sub cleanup {
	my ($self) = @_;

	syslog(LOG_DEBUG, "Removing entries older than $self->{max_age} days");

	$self->{cleanup_query}->execute($self->{max_age})
		or _terminate("DELETE failed: " . DBI->errstr . "\n");
	$self->{cleanup_query}->finish;
	$self->{db}->commit;
}

sub insert {
	my ($self, $value) = @_;
	my $query;

	_terminate('insert(): Missing argument $value') unless defined $value;

	if ($self->check($value)) {
		syslog(LOG_DEBUG, "Updating record for '$value' in database");
		$query = $self->{update_query};
	} else {
		syslog(LOG_DEBUG, "Inserting '$value' into database");
		$query = $self->{insert_query};
	}
	$query->execute($value)
		or _terminate("INSERT/UPDATE failed: " . DBI->errstr . "\n");
	$query->finish;
	$self->{db}->commit;
}

sub list {
	my ($self) = @_;
	my %records;
	my ($hash, $timestamp);

	syslog(LOG_DEBUG, 'Listing database entries');

	$self->{list_query}->execute()
		or _terminate("SELECT failed: " . DBI->errstr . "\n");
	$self->{list_query}->bind_columns(\$hash, \$timestamp)
		or _terminate("Binding result columns failed: " . DBI->errstr . "\n");
	while ($self->{list_query}->fetchrow_arrayref()) {
		$records{$hash} = $timestamp;
	}
	$self->{list_query}->finish;
	$self->{db}->rollback;

	return %records;
}

sub _terminate($) {
	syslog(LOG_ERR, $_[0]);
	die $_[0];
}

1;
